"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.core import CONFIG, CONN
from syndicate.core.helper import create_pool, unpack_kwargs
from syndicate.core.resources.helper import build_description_obj

_LOG = get_logger('syndicate.core.resources.cognito_identity_resource')
_COGNITO_IDENTITY_CONN = CONN.cognito_identity()


def cognito_resource_identifier(name, output=None):
    if output:
        # cognito currently is not located in different regions
        # process only first object
        pool_output = list(output.items())[0][1]
        # find id from the output
        return pool_output['description']['IdentityPoolId']
    return _COGNITO_IDENTITY_CONN.if_pool_exists_by_name(name)


def create_cognito_identity_pool(args):
    """ Create Cognito identity pool in pool in sub processes.

    :type args: list
    """
    return create_pool(_create_cognito_identity_pool_from_meta, args)


def describe_cognito_pool(name, meta, pool_id=None):
    if not pool_id:
        pool_id = _COGNITO_IDENTITY_CONN.if_pool_exists_by_name(name)
    if not pool_id:
        return
    response = _COGNITO_IDENTITY_CONN.describe_identity_pool(pool_id)
    arn = 'arn:aws:cognito-identity:{0}:{1}:identitypool/{2}'.format(
        CONFIG.region, CONFIG.account_id, pool_id)
    return {
        arn: build_description_obj(response, name, meta)
    }


@unpack_kwargs
def _create_cognito_identity_pool_from_meta(name, meta):
    """ Create Cognito identity pool for authentication.

    :type name: str
    :type meta: dict
    """
    pool_id = _COGNITO_IDENTITY_CONN.if_pool_exists_by_name(name)
    if pool_id:
        _LOG.warn('%s cognito identity pool exists.', name)
        return describe_cognito_pool(name=name, meta=meta, pool_id=pool_id)

    _LOG.info('Creating identity pool %s', name)
    open_id_provider_names = meta.get('open_id_providers', [])
    open_id_arns = ['arn:aws:iam::{0}:oidc-provider/{1}'.format(
        CONFIG.account_id, n) for n in open_id_provider_names]
    pool_id = _COGNITO_IDENTITY_CONN.create_identity_pool(
        pool_name=name, provider_name=meta.get('provider_name'),
        open_id_connect_provider_arns=open_id_arns)
    auth_role = meta.get('auth_role')
    unauth_role = meta.get('unauth_role')
    if auth_role or unauth_role:
        _COGNITO_IDENTITY_CONN.set_role(pool_id, auth_role, unauth_role)
    _LOG.info('Created cognito identity pool %s', pool_id)
    return describe_cognito_pool(name=name, meta=meta, pool_id=pool_id)


def remove_cognito_identity_pools(args):
    create_pool(_remove_cognito_identity_pool, args)


@unpack_kwargs
def _remove_cognito_identity_pool(arn, config):
    pool_id = config['description']['IdentityPoolId']
    try:
        _COGNITO_IDENTITY_CONN.remove_identity_pool(pool_id)
        _LOG.info('Cognito identity pool %s was removed', pool_id)
    except ClientError as e:
        if e.response['Error']['Code'] == 'ResourceNotFoundException':
            _LOG.warn('Cognito identity pool %s is not found', id)
        else:
            raise e



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from syndicate.core.constants import (API_GATEWAY_TYPE, CLOUD_WATCH_ALARM_TYPE,
                                      CLOUD_WATCH_RULE_TYPE, COGNITO_TYPE,
                                      DYNAMO_TABLE_TYPE, EBS_TYPE,
                                      EC2_INSTANCE_TYPE, IAM_POLICY,
                                      IAM_ROLE, KINESIS_STREAM_TYPE,
                                      LAMBDA_TYPE, S3_BUCKET_TYPE,
                                      SNS_PLATFORM_APPLICATION_TYPE,
                                      SNS_TOPIC_TYPE,
                                      SQS_QUEUE_TYPE, STATE_ACTIVITY_TYPE,
                                      STEP_FUNCTION_TYPE, LAMBDA_LAYER_TYPE)
from syndicate.core.resources.alarm_resource import (create_alarm,
                                                     describe_alarm,
                                                     remove_alarms)
from syndicate.core.resources.api_gateway_resource import (
    api_gateway_update_processor, api_resource_identifier, create_api_gateway,
    describe_api_resources, remove_api_gateways)
from syndicate.core.resources.cloud_watch_resource import (
    create_cloud_watch_rule, describe_rule, describe_rule_from_meta,
    remove_cloud_watch_rules)
from syndicate.core.resources.cognito_resource import (
    cognito_resource_identifier, create_cognito_identity_pool,
    describe_cognito_pool, remove_cognito_identity_pools)
from syndicate.core.resources.dynamo_db_resource import (create_tables_by_10,
                                                         describe_table,
                                                         remove_dynamodb_tables)
from syndicate.core.resources.ebs_resource import (create_ebs, describe_ebs,
                                                   remove_ebs_apps)
from syndicate.core.resources.ec2_resource import (create_ec2, describe_ec2,
                                                   remove_ec2_instances)
from syndicate.core.resources.iam_resource import (apply_policy_content,
                                                   apply_trusted_to_role,
                                                   create_policies,
                                                   create_roles,
                                                   describe_policy,
                                                   describe_role,
                                                   remove_policies,
                                                   remove_roles)
from syndicate.core.resources.kinesis_resource import (create_kinesis_stream,
                                                       describe_kinesis_stream,
                                                       remove_kinesis_streams)
from syndicate.core.resources.lambda_resource import (create_lambda,
                                                      describe_lambda,
                                                      remove_lambdas,
                                                      update_lambda,
                                                      create_lambda_layer,
                                                      remove_lambda_layers,
                                                      update_lambda_layer)
from syndicate.core.resources.s3_resource import (create_s3_bucket,
                                                  describe_bucket,
                                                  remove_buckets)
from syndicate.core.resources.sns_resource import (create_sns_application,
                                                   create_sns_topic,
                                                   describe_sns,
                                                   describe_sns_application,
                                                   describe_sns_application_from_meta,
                                                   describe_sns_from_meta,
                                                   remove_sns_application,
                                                   remove_sns_topics)
from syndicate.core.resources.sqs_resource import (create_sqs_queue,
                                                   describe_queue,
                                                   describe_queue_from_meta,
                                                   remove_queues)
from syndicate.core.resources.step_functions_resource import (
    create_activities, create_state_machine, describe_activity,
    describe_step_function, remove_activities, remove_state_machines)

CREATE_RESOURCE = {
    IAM_POLICY: create_policies,
    IAM_ROLE: create_roles,
    LAMBDA_TYPE: create_lambda,
    LAMBDA_LAYER_TYPE: create_lambda_layer,
    DYNAMO_TABLE_TYPE: create_tables_by_10,
    CLOUD_WATCH_RULE_TYPE: create_cloud_watch_rule,
    S3_BUCKET_TYPE: create_s3_bucket,
    API_GATEWAY_TYPE: create_api_gateway,
    COGNITO_TYPE: create_cognito_identity_pool,
    SNS_TOPIC_TYPE: create_sns_topic,
    SNS_PLATFORM_APPLICATION_TYPE: create_sns_application,
    SQS_QUEUE_TYPE: create_sqs_queue,
    CLOUD_WATCH_ALARM_TYPE: create_alarm,
    EBS_TYPE: create_ebs,
    STEP_FUNCTION_TYPE: create_state_machine,
    STATE_ACTIVITY_TYPE: create_activities,
    KINESIS_STREAM_TYPE: create_kinesis_stream,
    EC2_INSTANCE_TYPE: create_ec2
}

# 'ec2_instance' type is not supported
DESCRIBE_RESOURCE = {
    IAM_POLICY: describe_policy,
    IAM_ROLE: describe_role,
    LAMBDA_TYPE: describe_lambda,
    DYNAMO_TABLE_TYPE: describe_table,
    CLOUD_WATCH_RULE_TYPE: describe_rule_from_meta,
    S3_BUCKET_TYPE: describe_bucket,
    API_GATEWAY_TYPE: describe_api_resources,
    COGNITO_TYPE: describe_cognito_pool,
    SNS_TOPIC_TYPE: describe_sns_from_meta,
    SNS_PLATFORM_APPLICATION_TYPE: describe_sns_application_from_meta,
    SQS_QUEUE_TYPE: describe_queue_from_meta,
    CLOUD_WATCH_ALARM_TYPE: describe_alarm,
    EBS_TYPE: describe_ebs,
    STEP_FUNCTION_TYPE: describe_step_function,
    STATE_ACTIVITY_TYPE: describe_activity,
    KINESIS_STREAM_TYPE: describe_kinesis_stream,
}

REMOVE_RESOURCE = {
    CLOUD_WATCH_ALARM_TYPE: remove_alarms,
    API_GATEWAY_TYPE: remove_api_gateways,
    CLOUD_WATCH_RULE_TYPE: remove_cloud_watch_rules,
    COGNITO_TYPE: remove_cognito_identity_pools,
    DYNAMO_TABLE_TYPE: remove_dynamodb_tables,
    EBS_TYPE: remove_ebs_apps,
    EC2_INSTANCE_TYPE: remove_ec2_instances,
    IAM_ROLE: remove_roles,
    IAM_POLICY: remove_policies,
    KINESIS_STREAM_TYPE: remove_kinesis_streams,
    LAMBDA_LAYER_TYPE: remove_lambda_layers,
    LAMBDA_TYPE: remove_lambdas,
    S3_BUCKET_TYPE: remove_buckets,
    SNS_TOPIC_TYPE: remove_sns_topics,
    SNS_PLATFORM_APPLICATION_TYPE: remove_sns_application,
    SQS_QUEUE_TYPE: remove_queues,
    STEP_FUNCTION_TYPE: remove_state_machines,
    STATE_ACTIVITY_TYPE: remove_activities
}

UPDATE_RESOURCE = {
    LAMBDA_TYPE: update_lambda,
    LAMBDA_LAYER_TYPE: update_lambda_layer
}

RESOURCE_CONFIGURATION_PROCESSORS = {
    API_GATEWAY_TYPE: api_gateway_update_processor
}

RESOURCE_IDENTIFIER = {
    API_GATEWAY_TYPE: api_resource_identifier,
    COGNITO_TYPE: cognito_resource_identifier
}

APPLY_MAPPING = {
    IAM_ROLE: apply_trusted_to_role,
    IAM_POLICY: apply_policy_content
}



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json
import os

import click

from syndicate.core import CONFIG, CONF_PATH
from syndicate.core.build.artifact_processor import (RUNTIME_NODEJS,
                                                     assemble_artifacts,
                                                     RUNTIME_JAVA_8,
                                                     RUNTIME_PYTHON)
from syndicate.core.build.bundle_processor import (create_bundles_bucket,
                                                   load_bundle,
                                                   upload_bundle_to_s3,
                                                   if_bundle_exist)
from syndicate.core.build.deployment_processor import (
    continue_deployment_resources, create_deployment_resources,
    remove_deployment_resources, remove_failed_deploy_resources,
    update_lambdas, update_deployment_resources)
from syndicate.core.build.meta_processor import create_meta
from syndicate.core.conf.config_holder import (MVN_BUILD_TOOL_NAME,
                                               PYTHON_BUILD_TOOL_NAME,
                                               NODE_BUILD_TOOL_NAME)
from syndicate.core.helper import (check_required_param,
                                   create_bundle_callback,
                                   handle_futures_progress_bar,
                                   resolve_path_callback, timeit,
                                   verify_bundle_callback,
                                   verify_meta_bundle_callback,
                                   check_deploy_name_for_duplicates)


# TODO - command descriptions


@click.group(name='syndicate')
def syndicate():
    click.echo('Path to sdct.conf: ' + CONF_PATH)


# =============================================================================


@syndicate.command(name='clean')
@timeit
@click.option('--deploy_name', nargs=1, callback=check_required_param)
@click.option('--bundle_name', nargs=1, callback=check_required_param)
@click.option('--clean_only_types', multiple=True)
@click.option('--clean_only_resources', multiple=True)
@click.option('--clean_only_resources_path', nargs=1, type=str)
@click.option('--excluded_resources', multiple=True)
@click.option('--excluded_resources_path', nargs=1, type=str)
@click.option('--excluded_types', multiple=True)
@click.option('--rollback', is_flag=True)
def clean(deploy_name, bundle_name, clean_only_types, clean_only_resources,
          clean_only_resources_path, excluded_resources,
          excluded_resources_path, excluded_types, rollback):
    click.echo('Command clean')
    click.echo('Deploy name: %s' % deploy_name)
    if clean_only_types:
        click.echo('Clean only types: %s' % str(clean_only_types))
    if clean_only_resources:
        click.echo('Clean only resources : %s' % clean_only_resources)
    if clean_only_resources_path:
        click.echo(
            'Clean only resources path: %s' % clean_only_resources_path)
    if excluded_resources:
        click.echo('Excluded resources: %s' % str(excluded_resources))
    if excluded_resources_path:
        click.echo('Excluded resources path: %s' % excluded_resources_path)
    if excluded_types:
        click.echo('Excluded types: %s' % str(excluded_types))
    if clean_only_resources_path and os.path.exists(
            clean_only_resources_path):
        clean_resources_list = json.load(open(clean_only_resources_path))
        clean_only_resources = tuple(
            set(clean_only_resources + tuple(clean_resources_list)))
    if excluded_resources_path and os.path.exists(excluded_resources_path):
        excluded_resources_list = json.load(open(excluded_resources_path))
        excluded_resources = tuple(
            set(excluded_resources + tuple(excluded_resources_list)))
    if rollback:
        remove_failed_deploy_resources(deploy_name=deploy_name,
                                       bundle_name=bundle_name,
                                       clean_only_resources=clean_only_resources,
                                       clean_only_types=clean_only_types,
                                       excluded_resources=excluded_resources,
                                       excluded_types=excluded_types)
    else:
        remove_deployment_resources(deploy_name=deploy_name,
                                    bundle_name=bundle_name,
                                    clean_only_resources=clean_only_resources,
                                    clean_only_types=clean_only_types,
                                    excluded_resources=excluded_resources,
                                    excluded_types=excluded_types)
    click.echo('AWS resources were removed.')


# =============================================================================


@syndicate.command(name='assemble_java_mvn')
@timeit
@click.option('--bundle_name', nargs=1, callback=create_bundle_callback)
@click.option('--project_path', '-path', nargs=1,
              callback=resolve_path_callback)
def assemble_java_mvn(bundle_name, project_path):
    click.echo('Command compile java project path: %s' % project_path)
    assemble_artifacts(bundle_name=bundle_name,
                       project_path=project_path,
                       runtime=RUNTIME_JAVA_8)
    click.echo('Java artifacts were prepared successfully.')


@syndicate.command(name='assemble_python')
@timeit
@click.option('--bundle_name', nargs=1, callback=create_bundle_callback)
@click.option('--project_path', '-path', nargs=1,
              callback=resolve_path_callback)
def assemble_python(bundle_name, project_path):
    click.echo('Command assemble python: project_path: %s ' % project_path)
    assemble_artifacts(bundle_name=bundle_name,
                       project_path=project_path,
                       runtime=RUNTIME_PYTHON)
    click.echo('Python artifacts were prepared successfully.')


@syndicate.command(name='assemble_node')
@timeit
@click.option('--bundle_name', nargs=1, callback=create_bundle_callback)
@click.option('--project_path', '-path', nargs=1,
              callback=resolve_path_callback)
def assemble_node(bundle_name, project_path):
    click.echo('Command assemble node: project_path: %s ' % project_path)
    assemble_artifacts(bundle_name=bundle_name,
                       project_path=project_path,
                       runtime=RUNTIME_NODEJS)
    click.echo('NodeJS artifacts were prepared successfully.')


COMMAND_TO_BUILD_MAPPING = {
    MVN_BUILD_TOOL_NAME: assemble_java_mvn,
    PYTHON_BUILD_TOOL_NAME: assemble_python,
    NODE_BUILD_TOOL_NAME: assemble_node
}


@syndicate.command(name='build_artifacts')
@timeit
@click.option('--bundle_name', nargs=1, callback=create_bundle_callback)
@click.pass_context
def build_artifacts(ctx, bundle_name):
    click.echo('Building artifacts ...')
    if CONFIG.build_projects_mapping:
        for key, values in CONFIG.build_projects_mapping.items():
            for value in values:
                func = COMMAND_TO_BUILD_MAPPING.get(key)
                if func:
                    ctx.invoke(func, bundle_name=bundle_name,
                               project_path=value)
                else:
                    click.echo('Build tool is not supported: %s' % key)
    else:
        click.echo('Projects to be built are not found')


# =============================================================================


@syndicate.command(name='package_meta')
@timeit
@click.option('--bundle_name', nargs=1, callback=verify_bundle_callback)
def package_meta(bundle_name):
    click.echo('Package meta, bundle: %s' % bundle_name)
    create_meta(bundle_name)
    click.echo('Meta was configured successfully.')


# =============================================================================


@syndicate.command(name='create_deploy_target_bucket')
@timeit
def create_deploy_target_bucket():
    click.echo('Create deploy target sdk: %s' % CONFIG.deploy_target_bucket)
    create_bundles_bucket()
    click.echo('Deploy target bucket was created successfully')


@syndicate.command(name='upload_bundle')
@timeit
@click.option('--bundle_name', nargs=1, callback=verify_meta_bundle_callback)
@click.option('--force', is_flag=True)
def upload_bundle(bundle_name, force=False):
    click.echo('Upload bundle: %s' % bundle_name)
    if force:
        click.echo('Force upload')
    futures = upload_bundle_to_s3(bundle_name=bundle_name, force=force)
    handle_futures_progress_bar(futures)
    click.echo('Bundle was uploaded successfully')


@syndicate.command(name='copy_bundle')
@click.option('--bundle_name', nargs=1, callback=create_bundle_callback)
@click.option('--src_account_id', '-acc_id', nargs=1,
              callback=check_required_param)
@click.option('--src_bucket_region', '-r', nargs=1,
              callback=check_required_param)
@click.option('--src_bucket_name', '-bucket_name', nargs=1,
              callback=check_required_param)
@click.option('--role_name', '-role', nargs=1,
              callback=check_required_param)
@click.option('--force_upload', is_flag=True, default=False)
@timeit
@click.pass_context
def copy_bundle(ctx, bundle_name, src_account_id, src_bucket_region,
                src_bucket_name, role_name, force_upload):
    click.echo('Copy bundle: %s' % bundle_name)
    click.echo('Bundle name: %s' % bundle_name)
    click.echo('Source account id: %s' % src_account_id)
    click.echo('Source bucket region: %s' % src_bucket_region)
    click.echo('Source bucket name: %s' % src_bucket_name)
    futures = load_bundle(bundle_name, src_account_id, src_bucket_region,
                          src_bucket_name, role_name)
    handle_futures_progress_bar(futures)
    click.echo('Bundle was downloaded successfully')
    ctx.invoke(upload_bundle, bundle_name=bundle_name, force=force_upload)
    click.echo('Bundle was copied successfully')


# =============================================================================


@syndicate.command(name='build_bundle')
@click.option('--bundle_name', nargs=1, callback=check_required_param)
@click.option('--force_upload', is_flag=True, default=False)
@click.pass_context
@timeit
def build_bundle(ctx, bundle_name, force_upload):
    if if_bundle_exist(bundle_name=bundle_name) and not force_upload:
        click.echo('Bundle name \'{0}\' already exists '
                   'in deploy bucket. Please use another bundle '
                   'name or delete the bundle'.format(bundle_name))
        return
    ctx.invoke(build_artifacts, bundle_name=bundle_name)
    ctx.invoke(package_meta, bundle_name=bundle_name)
    ctx.invoke(upload_bundle, bundle_name=bundle_name, force=force_upload)


# =============================================================================


@syndicate.command(name='deploy')
@click.option('--deploy_name', nargs=1, callback=check_required_param)
@click.option('--bundle_name', nargs=1, callback=check_required_param)
@click.option('--deploy_only_types', multiple=True)
@click.option('--deploy_only_resources', multiple=True)
@click.option('--deploy_only_resources_path', nargs=1)
@click.option('--excluded_resources', multiple=True)
@click.option('--excluded_resources_path', nargs=1)
@click.option('--excluded_types', multiple=True)
@click.option('--continue_deploy', is_flag=True)
@click.option('--replace_output', nargs=1, is_flag=True, default=False)
@check_deploy_name_for_duplicates
@timeit
def deploy(deploy_name, bundle_name, deploy_only_types, deploy_only_resources,
           deploy_only_resources_path, excluded_resources,
           excluded_resources_path, excluded_types, continue_deploy,
           replace_output):
    click.echo('Command deploy backend')
    click.echo('Deploy name: %s' % deploy_name)
    if deploy_only_resources_path and os.path.exists(
            deploy_only_resources_path):
        deploy_resources_list = json.load(open(deploy_only_resources_path))
        deploy_only_resources = tuple(
            set(deploy_only_resources + tuple(deploy_resources_list)))
    if excluded_resources_path and os.path.exists(excluded_resources_path):
        excluded_resources_list = json.load(open(excluded_resources_path))
        excluded_resources = tuple(
            set(excluded_resources + tuple(excluded_resources_list)))
    if continue_deploy:
        deploy_success = continue_deployment_resources(deploy_name,
                                                       bundle_name,
                                                       deploy_only_resources,
                                                       deploy_only_types,
                                                       excluded_resources,
                                                       excluded_types,
                                                       replace_output)

    else:
        deploy_success = create_deployment_resources(deploy_name, bundle_name,
                                                     deploy_only_resources,
                                                     deploy_only_types,
                                                     excluded_resources,
                                                     excluded_types,
                                                     replace_output)
    click.echo('Backend resources were deployed{0}.'.format(
        '' if deploy_success else ' with errors. See deploy output file'))


# =============================================================================

@syndicate.command(name='update')
@click.option('--bundle_name', nargs=1, callback=check_required_param)
@click.option('--deploy_name', nargs=1, callback=check_required_param)
@click.option('--update_only_types', multiple=True)
@click.option('--update_only_resources', multiple=True)
@click.option('--update_only_resources_path', nargs=1)
@click.option('--replace_output', nargs=1, is_flag=True, default=False)
@check_deploy_name_for_duplicates
@timeit
def update(bundle_name, deploy_name, replace_output,
           update_only_resources,
           update_only_resources_path,
           update_only_types=[]):
    """
    Updates infrastructure from the provided bundle.
    :param bundle_name: name of the bundle to get updated meta
    :param deploy_name: name of the deploy
    :param update_only_resources: list of resources names to updated
    :param update_only_resources_path: path to a json file with list of
        resources names to update
    :param update_only_types: optional. List of a resources types to update.
    :param replace_output: flag. If True, existing output file will be replaced
    :return:
    """
    click.echo('Bundle name: {}'.format(bundle_name))
    if update_only_types:
        click.echo('Types to update: {}'.format(list(update_only_types)))
    if update_only_resources:
        click.echo('Resources to update: {}'.format(list(update_only_types)))
    if update_only_resources_path:
        click.echo('Path to list of resources to update: {}'.format(
            update_only_resources_path))

    if update_only_resources_path and os.path.exists(
            update_only_resources_path):
        update_resources_list = json.load(open(update_only_resources_path))
        update_only_resources = tuple(
            set(update_only_resources + tuple(update_resources_list)))
    success = update_deployment_resources(
        bundle_name=bundle_name,
        deploy_name=deploy_name,
        update_only_types=update_only_types,
        update_only_resources=update_only_resources,
        replace_output=replace_output)
    if success:
        return 'Update of resources has been successfully completed'
    return 'Something went wrong during resources update'


# =============================================================================

@syndicate.command(name='publish_lambda_version')
@click.option('--bundle_name', nargs=1, callback=check_required_param)
@click.option('--publish_only_lambdas', multiple=True)
@click.option('--publish_only_lambdas_path', nargs=1)
@click.option('--excluded_lambdas_resources', multiple=True)
@click.option('--excluded_lambdas_resources_path', nargs=1)
@click.option('--update_lambda_layers', is_flag=True)
@timeit
def publish_lambda_version(bundle_name,
                           publish_only_lambdas, publish_only_lambdas_path,
                           excluded_lambdas_resources,
                           excluded_lambdas_resources_path,
                           update_lambda_layers):
    click.secho('The command \'publish_lambda_version\' is deprecated. '
                'Please, use \'syndicate update\' instead',
                fg='black', bg='red')
    click.echo('Bundle name: %s' % bundle_name)
    if publish_only_lambdas_path and os.path.exists(
            publish_only_lambdas_path):
        update_lambdas_list = json.load(open(publish_only_lambdas_path))
        publish_only_lambdas = tuple(
            set(publish_only_lambdas + tuple(update_lambdas_list)))
    if excluded_lambdas_resources_path and os.path.exists(
            excluded_lambdas_resources_path):
        excluded_lambdas_list = json.load(
            open(excluded_lambdas_resources_path))
        excluded_lambdas_resources = tuple(
            set(excluded_lambdas_resources + tuple(excluded_lambdas_list)))
    update_lambdas(bundle_name=bundle_name,
                   publish_only_lambdas=publish_only_lambdas,
                   excluded_lambdas_resources=excluded_lambdas_resources,
                   update_lambda_layers=update_lambda_layers)
    click.echo('Lambda versions were published.')



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import time

from botocore.exceptions import ClientError
from syndicate.commons.log_helper import get_logger
from syndicate.core import CONFIG, CONN
from syndicate.core.helper import create_pool, unpack_kwargs
from syndicate.core.resources.helper import build_description_obj

_LOG = get_logger('core.resources.sqs_resource')
AVAILABLE_REGIONS = ['us-east-2', 'us-east-1', 'us-west-1', 'us-west-2',
                     'ap-south-1', 'ap-northeast-2', 'ap-southeast-1',
                     'ap-southeast-2', 'ap-northeast-1', 'ca-central-1',
                     'eu-central-1', 'eu-west-2', 'sa-east-1', 'eu-west-1']
FIFO_REGIONS = ['us-east-1', 'us-east-2', 'us-west-2', 'eu-west-1']


def create_sqs_queue(args):
    return create_pool(_create_sqs_queue_from_meta, args)


def describe_queue(queue_url, name, meta, resource_name, region):
    response = CONN.sqs(region).get_queue_attributes(queue_url)
    arn = _build_queue_arn(resource_name=resource_name, region=region)
    return {arn: build_description_obj(response, name, meta)}


def describe_queue_from_meta(name, meta):
    region = meta.get('region', CONFIG.region)
    is_fifo = meta.get('fifo_queue', False)
    resource_name = _build_resource_name(is_fifo, name)
    queue_url = CONN.sqs(region).get_queue_url(resource_name,
                                               CONFIG.account_id)
    if not queue_url:
        return {}
    response = CONN.sqs(region).get_queue_attributes(queue_url)
    return {
        _build_queue_arn(resource_name, region): build_description_obj(
            response, name, meta)
    }


def remove_queues(args):
    create_pool(_remove_queue, args)
    # wait to remove all queues
    if args:
        time.sleep(60)


@unpack_kwargs
def _remove_queue(arn, config):
    region = arn.split(':')[3]
    queue_name = config['resource_name']
    resource_meta = config['resource_meta']
    try:
        is_fifo = resource_meta.get('fifo_queue', False)
        resource_name = _build_resource_name(is_fifo, queue_name)
        queue_url = CONN.sqs(region).get_queue_url(resource_name,
                                                   CONFIG.account_id)
        if queue_url:
            CONN.sqs(region).delete_queue(queue_url)
            _LOG.info('SQS queue %s was removed.', queue_name)
        else:
            _LOG.warn('SQS queue %s is not found', queue_name)
    except ClientError as e:
        exception_type = e.response['Error']['Code']
        if exception_type == 'ResourceNotFoundException':
            _LOG.warn('SQS queue %s is not found', queue_name)
        else:
            raise e


@unpack_kwargs
def _create_sqs_queue_from_meta(name, meta):
    region = meta.get('region', CONFIG.region)
    is_fifo = meta.get('fifo_queue', False)
    resource_name = _build_resource_name(is_fifo, name)
    queue_url = CONN.sqs(region).get_queue_url(resource_name,
                                               CONFIG.account_id)
    if queue_url:
        _LOG.warn('SQS queue %s exists.', name)
        return describe_queue(queue_url, name, meta, resource_name, region)
    delay_sec = meta.get('delay_seconds')
    max_mes_size = meta.get('maximum_message_size')
    mes_ret_period = meta.get('message_retention_period')
    policy = meta.get('policy')
    recieve_mes_wait_sec = meta.get('receive_message_wait_time_seconds')
    redrive_policy = meta.get('redrive_policy')
    vis_timeout = meta.get('visibility_timeout')
    kms_master_key_id = meta.get('kms_master_key_id')
    kms_data_reuse_period = meta.get('kms_data_key_reuse_period_seconds')
    if is_fifo and region not in FIFO_REGIONS:
        raise AssertionError('FIFO queue is not available in {0}.',
                             region)
    content_deduplication = meta.get('content_based_deduplication')
    params = dict(queue_name=resource_name,
                  delay_seconds=delay_sec,
                  maximum_message_size=max_mes_size,
                  message_retention_period=mes_ret_period,
                  policy=policy,
                  receive_message_wait_time_seconds=recieve_mes_wait_sec,
                  redrive_policy=redrive_policy,
                  visibility_timeout=vis_timeout,
                  kms_master_key_id=kms_master_key_id,
                  kms_data_key_reuse_period_seconds=kms_data_reuse_period,
                  fifo_queue=is_fifo,
                  content_based_deduplication=content_deduplication)
    queue_url = CONN.sqs(region).create_queue(**params)['QueueUrl']
    _LOG.info('Created SQS queue %s.', name)
    return describe_queue(queue_url, name, meta, resource_name, region)


def _build_resource_name(is_fifo, name):
    resource_name = name
    if is_fifo:
        resource_name += '.fifo'
    return resource_name


def _build_queue_arn(resource_name, region):
    arn = 'arn:aws:sqs:{0}:{1}:{2}'.format(region, CONFIG.account_id,
                                           resource_name)
    return arn



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os

from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.connection import ConnectionProvider
from syndicate.connection.sts_connection import STSConnection
from syndicate.core.conf.config_holder import ConfigHolder

_LOG = get_logger('deployment.__init__')

S3_PATH_NAME = 's3_path'
SESSION_TOKEN = 'aws_session_token'
SECRET_KEY = 'aws_secret_access_key'
ACCESS_KEY = 'aws_access_key_id'


def exception_handler(exception_type, exception, traceback):
    print(exception)


# sys.excepthook = exception_handler

# suppress botocore warnings

try:
    CONF_PATH = os.environ['SDCT_CONF']
except KeyError:
    raise AssertionError('Environment variable SDCT_CONF is not set! '
                         'Please verify that you configured '
                         'framework correctly.')

# CONF VARS ===================================================================
CONFIG = ConfigHolder(CONF_PATH)

sts = STSConnection(CONFIG.region, CONFIG.aws_access_key_id,
                    CONFIG.aws_secret_access_key)


def _ready_to_assume():
    return CONFIG.access_role and CONFIG.aws_access_key_id and \
           CONFIG.aws_secret_access_key


def _ready_to_use_creds():
    return not CONFIG.access_role and CONFIG.aws_access_key_id and \
           CONFIG.aws_secret_access_key


try:
    CREDENTIALS = {
        'region': CONFIG.region
    }
    if _ready_to_assume():
        _LOG.debug('Starting to assume role ...')
        # get credentials for 12 hours
        temp_credentials = sts.get_temp_credentials(CONFIG.access_role,
                                                    CONFIG.account_id,
                                                    43200)
        _LOG.debug('Role %s is assumed successfully' % CONFIG.access_role)
        CREDENTIALS[ACCESS_KEY] = temp_credentials['AccessKeyId']
        CREDENTIALS[SECRET_KEY] = temp_credentials['SecretAccessKey']
        CREDENTIALS[SESSION_TOKEN] = temp_credentials['SessionToken']
    elif _ready_to_use_creds():
        _LOG.debug('Credentials access')
        CREDENTIALS[ACCESS_KEY] = CONFIG.aws_access_key_id
        CREDENTIALS[SECRET_KEY] = CONFIG.aws_secret_access_key
    CONN = ConnectionProvider(CREDENTIALS)
except ClientError:
    raise AssertionError('Cannot assume {0} role. '
                         'Please verify that you have configured '
                         'the role correctly.'.format(CONFIG.access_role))



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from time import time
from uuid import uuid1

from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.core import CONFIG, CONN
from syndicate.core.build.meta_processor import S3_PATH_NAME
from syndicate.core.helper import create_pool, unpack_kwargs
from syndicate.core.resources.helper import build_description_obj

_LOG = get_logger('syndicate.core.resources.ebs_resource')

_EC2_CONN = CONN.ec2()
_IAM_CONN = CONN.iam()
_EBS_CONN = CONN.beanstalk()


def describe_ebs(name, meta, response=None):
    arn = 'arn:aws:elasticbeanstalk:{0}:{1}:application/{2}'.format(
        CONFIG.region, CONFIG.account_id, name)
    if not response:
        response = _EBS_CONN.describe_applications([name])
    return {
        arn: build_description_obj(response, name, meta)
    }


def create_ebs(args):
    return create_pool(_create_ebs_app_env_from_meta, args)


@unpack_kwargs
def _create_ebs_app_env_from_meta(name, meta):
    response = _EBS_CONN.describe_applications([name])
    if response:
        _LOG.warn('%s EBS app exists.', name)
        return describe_ebs(name, meta, response[0])

    env_settings = meta['env_settings']
    topic_name = meta.get('notification_topic')
    # check topic exists
    if topic_name:
        topic_arn = CONN.sns().get_topic_arn(topic_name)
        if topic_arn:
            env_settings.append({
                "OptionName": "Notification Topic ARN",
                "Namespace": "aws:elasticbeanstalk:sns:topics",
                "Value": "{0}".format(topic_arn)
            })
        else:
            raise AssertionError('Cant find notification '
                                 'topic {0} for EBS.'.format(topic_name))
    # check key pair exists
    key_pair_name = meta['ec2_key_pair']
    if _EC2_CONN.if_key_pair_exists(key_pair_name):
        env_settings.append({
            "OptionName": "KeyName",
            "ResourceName": "AWSEBAutoScalingLaunchConfiguration",
            "Namespace": "aws:cloudformation:template:resource:property",
            "Value": key_pair_name
        })
    else:
        raise AssertionError('Specified key pair '
                             'does not exist: {0}.'.format(key_pair_name))
    # check ec2 role exists
    iam_role = meta['ec2_role']
    if _IAM_CONN.check_if_role_exists(iam_role):
        env_settings.append({
            "OptionName": "IamInstanceProfile",
            "ResourceName": "AWSEBAutoScalingLaunchConfiguration",
            "Namespace": "aws:autoscaling:launchconfiguration",
            "Value": iam_role
        })
    else:
        raise AssertionError(
            'Specified iam role does not exist: {0}.'.format(iam_role))
    # check service role exists
    iam_role = meta['ebs_service_role']
    if _IAM_CONN.check_if_role_exists(iam_role):
        env_settings.append({
            "OptionName": "ServiceRole",
            "Namespace": "aws:elasticbeanstalk:environment",
            "Value": iam_role
        })
    else:
        raise AssertionError('Specified iam role '
                             'does not exist: {0}.'.format(iam_role))
    image_id = meta.get('image_id')
    if image_id:
        env_settings.append({
            "OptionName": "ImageId",
            "ResourceName": "AWSEBAutoScalingLaunchConfiguration",
            "Namespace": "aws:autoscaling:launchconfiguration",
            "Value": image_id
        })
    else:
        _LOG.warn('Image id is not specified.')
    # check that desired solution stack exists
    stack = meta['stack']
    available_stacks = _EBS_CONN.describe_available_solutions_stack_names()
    if stack not in available_stacks:
        raise AssertionError('No solution stack named {0} found.'
                             ' Available:\n{1}'.format(stack,
                                                       available_stacks))
    vpc_id = next(
        (option for option in env_settings if option['OptionName'] == 'VPCId'),
        None)
    if not vpc_id:
        vpc_id = _EC2_CONN.get_default_vpc_id()
        _LOG.info('Default vpc id %s', vpc_id)
        if vpc_id:
            _LOG.debug('Will use vpc %s', vpc_id)
            subnets = _EC2_CONN.list_subnets(filters=[{
                'Name': 'vpc-id',
                'Values': [vpc_id]
            }])
            _LOG.debug('Found subnets for %s vpc: %s', vpc_id, subnets)
            if subnets:
                _LOG.info('Will attach default %s vpc to env', vpc_id)
                _add_subnets_info(env_settings, subnets, vpc_id)
            sg_id = _EC2_CONN.get_sg_id(group_name='default', vpc_id=vpc_id)
            if sg_id:
                _LOG.debug('Found default sg with id %s', sg_id)
                env_settings.append({
                    "OptionName": "SecurityGroups",
                    "Namespace": "aws:autoscaling:launchconfiguration",
                    "Value": sg_id
                })

    env_name = meta["env_name"] + str(int(time()))

    start = time()
    end = start + 180
    while end > time():
        describe_app_result = _EBS_CONN.describe_applications([name])
        if not describe_app_result:
            break

    # create APP
    response = _EBS_CONN.create_application(name)
    _LOG.info('Created EBS app %s.', name)
    # create ENV
    _EBS_CONN.create_environment(app_name=name,
                                 env_name=env_name,
                                 option_settings=env_settings,
                                 tier=meta['tier'],
                                 solution_stack_name=stack)
    key = meta[S3_PATH_NAME]
    if not CONN.s3().is_file_exists(CONFIG.deploy_target_bucket, key):
        raise AssertionError('Deployment package does not exist in '
                             '{0} bucket'.format(CONFIG.deploy_target_bucket))

    # create VERSION
    version_label = env_name + str(uuid1())
    _EBS_CONN.create_app_version(app_name=name, version_label=version_label,
                                 s3_bucket=CONFIG.deploy_target_bucket,
                                 s3_key=key)
    _LOG.debug('Waiting for beanstalk env %s', env_name)
    # wait for env creation
    start = time()
    status = {}
    end = start + 360  # end in 6 min
    while end > time():
        status = _EBS_CONN.describe_environment_health(env_name=env_name,
                                                       attr_names=['Status'])
        if status['Status'] == 'Ready':
            _LOG.info('Launching env took %s.', time() - start)
            break
    if status['Status'] != 'Ready':
        _LOG.error('Env status: %s. Failed to create env.', status)
    # deploy new app version
    _EBS_CONN.deploy_env_version(name, env_name, version_label)
    _LOG.info('Created environment for %s.', name)
    return describe_ebs(name, meta, response)


def _add_subnets_info(env_settings, subnets, vpc_id):
    env_settings.append({
        "OptionName": "VPCId",
        "Namespace": "aws:ec2:vpc",
        "Value": vpc_id
    })
    subnets = ",".join(subnet['SubnetId'] for subnet in subnets)
    env_settings.append({
        "OptionName": "Subnets",
        "Namespace": "aws:ec2:vpc",
        "Value": subnets
    })


def remove_ebs_apps(args):
    create_pool(_remove_ebs_app, args)


@unpack_kwargs
def _remove_ebs_app(arn, config):
    app_name = config['resource_name']
    try:
        _EBS_CONN.remove_app(app_name)
        _LOG.info('EBS app %s was removed.', app_name)
    except ClientError as e:
        if e.response['Error']['Code'] == 'ResourceNotFoundException':
            _LOG.warn('EBS app %s is not found', app_name)
        else:
            raise e



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from syndicate.commons.log_helper import get_logger
from syndicate.core import CONN, ClientError
from syndicate.core.helper import create_pool, unpack_kwargs
from syndicate.core.resources.helper import build_description_obj, chunks

_LOG = get_logger('syndicate.core.resources.s3_resource')
_S3_CONN = CONN.s3()


def create_s3_bucket(args):
    return create_pool(_create_s3_bucket_from_meta, args)


def describe_bucket(name, meta):
    arn = 'arn:aws:s3:::{0}'.format(name)
    acl_response = _S3_CONN.get_bucket_acl(name)
    location_response = _S3_CONN.get_bucket_location(name)
    bucket_policy = _S3_CONN.get_bucket_policy(name)
    response = {
        'bucket_acl': acl_response,
        'location': location_response,
    }
    if bucket_policy:
        response['policy'] = bucket_policy
    return {
        arn: build_description_obj(response, name, meta)
    }


@unpack_kwargs
def _create_s3_bucket_from_meta(name, meta):
    if _S3_CONN.is_bucket_exists(name):
        _LOG.warn('{0} bucket exists.'.format(name))
        return describe_bucket(name, meta)
    _S3_CONN.create_bucket(name, meta.get('acl'), meta.get('location'))
    policy = meta.get('policy')
    if policy:
        _S3_CONN.add_bucket_policy(name, policy)
        _LOG.debug('Policy on {0} S3 bucket is set up.'.format(name))
    _LOG.info('Created S3 bucket {0}.'.format(name))
    rules = meta.get('LifecycleConfiguration')
    if rules:
        _S3_CONN.add_bucket_rule(name, rules)
        _LOG.debug('Rules on {0} S3 bucket are set up.'.format(name))
    cors_configuration = meta.get('cors')
    if cors_configuration:
        _S3_CONN.put_cors(bucket_name=name, rules=cors_configuration)
    return describe_bucket(name, meta)


def _delete_objects(bucket_name, keys):
    response = _S3_CONN.delete_objects(bucket_name, keys)
    errors = response.get('Errors')
    if errors:
        error_keys = [{
            'Key': i['Key'],
            'VersionId': i['VersionId']
        } for i in errors]
        return error_keys
    else:
        return []


def remove_buckets(args):
    create_pool(_remove_bucket, args)


@unpack_kwargs
def _remove_bucket(arn, config):
    bucket_name = config['resource_name']
    try:
        errors = []
        keys = _S3_CONN.list_object_versions(bucket_name)
        if keys:
            for s3_keys in chunks(keys, 1000):
                errors.extend(_delete_objects(bucket_name, s3_keys))

        markers = _S3_CONN.list_object_markers(bucket_name)
        if markers:
            for s3_markers in chunks(markers, 1000):
                errors.extend(_delete_objects(bucket_name, s3_markers))

        if errors:
            raise AssertionError('Error occurred while deleting S3 objects'
                                 ' from {0} bucket. Not deleted keys: '
                                 '{1}'.format(bucket_name, str(errors)))
        else:
            _S3_CONN.delete_bucket(bucket_name)
            _LOG.info('S3 bucket {0} was removed.'.format(bucket_name))
    except ClientError as e:
        if e.response['Error']['Code'] == 'NoSuchBucket':
            _LOG.warn('S3 bucket {0} is not found'.format(bucket_name))
        else:
            raise e



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os
from time import sleep

from syndicate.commons.log_helper import get_logger
from syndicate.core import CONFIG, CONF_PATH, CONN, ClientError
from syndicate.core.helper import create_pool, unpack_kwargs
from syndicate.core.resources.helper import build_description_obj, chunks

_LOG = get_logger('syndicate.core.resources.ec2_resource')
_EC2_CONN = CONN.ec2()
_IAM_CONN = CONN.iam()


def describe_ec2(name, meta, response=None):
    if not response:
        response = _EC2_CONN.describe_instances()
    arn = 'arn:aws:ec2:{0}:{1}:instance/{2}'.format(CONFIG.region,
                                                    CONFIG.account_id,
                                                    response['InstanceId'])
    describe_response = _EC2_CONN.describe_instances(
        [
            {
                'Name': 'instance-id',
                'Values': [response['InstanceId']]
            }
        ]
    )
    response['NetworkInterfaces'] = describe_response
    return {
        arn: build_description_obj(response, name, meta)
    }


def create_ec2(args):
    return create_pool(_create_ec2_from_meta, args)


@unpack_kwargs
def _create_ec2_from_meta(name, meta):
    # checking required parameters
    image_id = meta['image_id']
    image_data = _EC2_CONN.describe_image(image_id=image_id)
    if not image_data:
        raise AssertionError('Image id {0} is invalid'.format(image_id))

    instance_type = meta.get('instance_type')
    if not instance_type:
        raise AssertionError('Instance type must be specified')

    key_name = meta.get('key_name')
    if not _EC2_CONN.if_key_pair_exists(key_name):
        raise AssertionError('There is no key pair with name: {0}'
                             .format(key_name))

    availability_zone = meta.get('availability_zone')
    subnet = meta.get('subnet_id')
    if availability_zone:
        subnet_filter = {
            'Name': 'availabilityZone',
            'Values': [availability_zone]
        }
        subnet_list = _EC2_CONN.list_subnets(filters=[subnet_filter])
        if subnet and subnet not in \
                [subnet_ids['SubnetId'] for subnet_ids in subnet_list]:
            raise AssertionError('There is no available Subnets with name {0} '
                                 'in Availability Zone {1}.'
                                 .format(subnet, availability_zone))
        if availability_zone not in _EC2_CONN.get_azs():
            raise AssertionError('There is no Availability Zone with name: {0}'
                                 .format(availability_zone))

    security_groups_names = meta.get('security_group_names')
    if security_groups_names:
        sg_meta = _EC2_CONN.describe_security_groups(security_groups_names)
        described_sec_groups_names = [security_group['GroupName']
                                      for security_group in sg_meta]
        for security_group_name in security_groups_names:
            if security_group_name not in described_sec_groups_names:
                raise AssertionError('Security group {0} does not exist'
                                     .format(security_group_name))

    # checking optional parameters
    user_data_file_name = meta.get('userdata_file')
    user_data_content = None
    if user_data_file_name:
        user_data_location = os.path.join(CONF_PATH, user_data_file_name)
        if not os.path.isfile(user_data_location):
            _LOG.warn('There is no user data {0} found by path {1}. '
                      .format(user_data_file_name, CONF_PATH))
        else:
            with open(user_data_location, 'r') as userdata_file:
                user_data_content = userdata_file.read()

    # describing instance profile by iam role name
    iam_role_name = meta.get('iam_role')
    iam_instance_profile_object = None
    if iam_role_name:
        instance_profiles = _IAM_CONN.get_instance_profiles_for_role(
            role_name=iam_role_name)
        if instance_profiles:
            iam_profile_meta = instance_profiles[0]
            iam_instance_profile_arn = iam_profile_meta['Arn']
            iam_instance_profile_object = {
                'Arn': iam_instance_profile_arn
            }

    # launching instance
    response = _EC2_CONN.launch_instance(
        name=name,
        image_id=image_id,
        instance_type=instance_type,
        key_name=key_name,
        tags_list=meta.get('tags_list'),
        security_groups_names=meta.get('security_group_names'),
        security_group_ids=meta.get('security_group_ids'),
        user_data=user_data_content,
        iam_instance_profile=iam_instance_profile_object,
        subnet_id=meta.get('subnet_id'),
        availability_zone=availability_zone
    )

    if meta.get('disableApiTermination'):
        disable_api_termination = meta.get('disableApiTermination')
        _LOG.debug('Found disableApiTermination '
                   'property: {0}'.format(disable_api_termination))
        if str(disable_api_termination).lower() == 'true':
            _EC2_CONN.modify_instance_attribute(
                InstanceId=response['InstanceId'],
                DisableApiTermination={
                    'Value': True
                }
            )
    _LOG.info('Created EC2 instance %s. '
              'Waiting for instance network interfaces configuring.', name)
    sleep(30)  # time for vm to become running
    return describe_ec2(name, meta, response)


def remove_ec2_instances(args):
    create_pool(remove_instance_list, chunks(args, 1000))


def remove_instance_list(*instance_list):
    instance_ids = [x['config']['description']['InstanceId'] for x in
                    instance_list[0]]
    existing_instances_list = []
    for instance_id in instance_ids:
        try:
            _EC2_CONN.modify_instance_attribute(
                InstanceId=instance_id,
                DisableApiTermination={'Value': False}
            )
            existing_instances_list.append(instance_id)
        except ClientError as e:
            if 'InvalidInstanceID.NotFound' in str(e):
                _LOG.warn('Instance %s does not exist', instance_id)
            else:
                raise e

    if existing_instances_list:
        _EC2_CONN.terminate_instances(instance_ids=existing_instances_list)
    _LOG.info('EC2 instances %s were removed.', str(existing_instances_list))



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.core import CONN
from syndicate.core.helper import create_pool, unpack_kwargs
from syndicate.core.resources.helper import (build_description_obj, chunks,
                                             validate_params)

_LOG = get_logger('syndicate.core.resources.alarm_resource')
_CW_METRIC = CONN.cw_metric()


def create_alarm(args):
    """ Create alarm in pool in sub processes.

    :type args: list
    """
    return create_pool(_create_alarm_from_meta, args)


def describe_alarm(name, meta):
    response = _CW_METRIC.describe_alarms([name])[0]
    arn = response['AlarmArn']
    return {
        arn: build_description_obj(response, name, meta)
    }


@unpack_kwargs
def _create_alarm_from_meta(name, meta):
    """ Create alarm resource in AWS Cloud via meta description.

    :type name: str
    :type meta: dict
    """
    required_parameters = ['metric_name', 'namespace', 'period', 'threshold',
                           'evaluation_periods', 'comparison_operator',
                           'statistic']
    validate_params(name, meta, required_parameters)

    if _CW_METRIC.is_alarm_exists(name):
        _LOG.warn('%s alarm exists.', name)
        return describe_alarm(name, meta)

    params = dict(alarm_name=name, metric_name=meta['metric_name'],
                  namespace=meta['namespace'], period=meta['period'],
                  evaluation_periods=meta['evaluation_periods'],
                  threshold=meta['threshold'], statistic=meta['statistic'],
                  comparison_operator=meta['comparison_operator'])

    sns_topics = meta.get('sns_topics')
    sns_topic_arns = []
    if sns_topics:
        for each in sns_topics:
            arn = CONN.sns().get_topic_arn(each)
            sns_topic_arns.append(arn)
        if sns_topic_arns:
            params['alarm_actions'] = sns_topic_arns

    _CW_METRIC.put_metric_alarm(**params)
    _LOG.info('Created alarm {0}.'.format(name))
    return describe_alarm(name, meta)


def remove_alarms(args):
    create_pool(remove_alarm_list, chunks(args, 100))


def remove_alarm_list(*alarm_list):
    alarm_names = [x['config']['resource_name'] for x in alarm_list[0]]
    try:
        _CW_METRIC.remove_alarms(alarm_names=alarm_names)
        _LOG.info('Alarms %s were removed.', str(alarm_names))
    except ClientError as e:
        if e.response['Error']['Code'] == 'ResourceNotFoundException':
            _LOG.warn('Alarms %s are not found', str(alarm_names))
        else:
            raise e



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import time

from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import retry
from syndicate.core import CONFIG, CONN
from syndicate.core.build.meta_processor import S3_PATH_NAME
from syndicate.core.helper import (create_pool, unpack_kwargs)
from syndicate.core.resources.helper import (build_description_obj,
                                             validate_params)
from syndicate.core.resources.sns_resource import (
    create_sns_subscription_for_lambda)

_LOG = get_logger('syndicate.core.resources.lambda_resource')
_LAMBDA_CONN = CONN.lambda_conn()
_S3_CONN = CONN.s3()
_CW_LOGS_CONN = CONN.cw_logs()


def create_lambda(args):
    """ Create lambdas in pool in sub processes.

    :type args: list
    """
    return create_pool(_create_lambda_from_meta, args)


def update_lambda(args):
    return create_pool(_update_lambda, args)


def update_lambda_layer(args):
    return create_pool(create_lambda_layer_from_meta, args)


def describe_lambda(name, meta, response=None):
    if not response:
        response = _LAMBDA_CONN.get_function(lambda_name=name)
    arn = build_lambda_arn_with_alias(response, meta.get('alias'))

    del response['Configuration']['FunctionArn']
    return {
        arn: build_description_obj(response, name, meta)
    }


def build_lambda_arn(name):
    arn = 'arn:aws:lambda:{0}:{1}:function:{2}'.format(CONFIG.region,
                                                       CONFIG.account_id, name)
    return arn


def resolve_lambda_arn_by_version_and_alias(name, version, alias):
    if version or alias:
        lambda_response = _LAMBDA_CONN.get_function(name, version)
        return build_lambda_arn_with_alias(lambda_response, alias)
    else:
        return _LAMBDA_CONN.get_function(name)['Configuration']['FunctionArn']


def build_lambda_arn_with_alias(response, alias=None):
    name = response['Configuration']['FunctionName']
    l_arn = build_lambda_arn(name=name)
    version = response['Configuration']['Version']
    arn = '{0}:{1}'.format(l_arn, version)
    # override version if alias exists
    if alias:
        arn = '{0}:{1}'.format(l_arn, alias)
    return arn


@unpack_kwargs
def _create_lambda_from_meta(name, meta):
    req_params = ['iam_role_name', 'runtime', 'memory', 'timeout', 'func_name']

    # Lambda configuration
    validate_params(name, meta, req_params)

    key = meta[S3_PATH_NAME]
    if not _S3_CONN.is_file_exists(CONFIG.deploy_target_bucket, key):
        raise AssertionError('Deployment package %s does not exist '
                             'in %s bucket', key, CONFIG.deploy_target_bucket)

    response = _LAMBDA_CONN.get_function(name)
    if response:
        _LOG.warn('%s lambda exists.', name)
        return describe_lambda(name, meta, response)

    role_name = meta['iam_role_name']
    role_arn = CONN.iam().check_if_role_exists(role_name)
    if not role_arn:
        raise AssertionError('Role {0} does not exist.'.format(role_name))

    dl_type = meta.get('dl_resource_type')
    if dl_type:
        dl_type = dl_type.lower()
    dl_name = meta.get('dl_resource_name')

    dl_target_arn = 'arn:aws:{0}:{1}:{2}:{3}'.format(dl_type,
                                                     CONFIG.region,
                                                     CONFIG.account_id,
                                                     dl_name) if dl_type and dl_name else None

    publish_version = meta.get('publish_version', False)
    lambda_layers_arns = []
    layer_meta = meta.get('layers')
    if layer_meta:
        for layer_name in layer_meta:
            layer_arn = _LAMBDA_CONN.get_lambda_layer_arn(layer_name)
            if not layer_arn:
                raise AssertionError(
                    'Lambda layer {} is absent in your account!')
            lambda_layers_arns.append(layer_arn)

    _LOG.debug('Creating lambda %s', name)
    _LAMBDA_CONN.create_lambda(
        lambda_name=name,
        func_name=meta['func_name'],
        role=role_arn,
        runtime=meta['runtime'].lower(),
        memory=meta['memory'],
        timeout=meta['timeout'],
        s3_bucket=CONFIG.deploy_target_bucket,
        s3_key=key,
        env_vars=meta.get('env_variables'),
        vpc_sub_nets=meta.get('subnet_ids'),
        vpc_security_group=meta.get('security_group_ids'),
        dl_target_arn=dl_target_arn,
        tracing_mode=meta.get('tracing_mode'),
        publish_version=publish_version,
        layers=lambda_layers_arns
    )
    _LOG.debug('Lambda created %s', name)
    # AWS sometimes returns None after function creation, needs for stability
    time.sleep(10)
    response = __describe_lambda_by_version(
        name) if publish_version else _LAMBDA_CONN.get_function(name)
    version = response['Configuration']['Version']
    con_exec = meta.get('concurrent_executions')
    if con_exec:
        _LOG.debug('Going to set up concurrency executions')
        unresolved_exec = _LAMBDA_CONN.get_unresolved_concurrent_executions()
        if con_exec <= unresolved_exec:
            _LAMBDA_CONN.put_function_concurrency(
                function_name=name,
                concurrent_executions=con_exec)
            _LOG.debug('Concurrency is enabled for %s lambda', name)
        else:
            _LOG.warn(
                'Account does not have any unresolved executions.'
                ' Current size - %s', unresolved_exec)

    # enabling aliases
    # aliases can be enabled only and for $LATEST
    alias = meta.get('alias')
    if alias:
        _LOG.debug('Creating alias')
        _LOG.debug(_LAMBDA_CONN.create_alias(function_name=name,
                                             name=alias, version=version))

    arn = build_lambda_arn_with_alias(response,
                                      alias) if publish_version or alias else \
        response['Configuration']['FunctionArn']
    _LOG.debug('arn value: ' + str(arn))

    if meta.get('event_sources'):
        for trigger_meta in meta.get('event_sources'):
            trigger_type = trigger_meta['resource_type']
            func = CREATE_TRIGGER[trigger_type]
            func(name, arn, role_name, trigger_meta)
    _LOG.info('Lambda %s has been successfully configured', name)
    return describe_lambda(name, meta, response)


@unpack_kwargs
def _update_lambda(name, meta, context):
    _LOG.info('Updating lambda: {0}'.format(name))
    req_params = ['runtime', 'memory', 'timeout', 'func_name']

    validate_params(name, meta, req_params)

    key = meta[S3_PATH_NAME]
    if not _S3_CONN.is_file_exists(CONFIG.deploy_target_bucket, key):
        raise AssertionError(
            'Deployment package {0} does not exist '
            'in {1} bucket'.format(key, CONFIG.deploy_target_bucket))

    response = _LAMBDA_CONN.get_function(name)
    if not response:
        raise AssertionError('{0} lambda does not exist.'.format(name))

    publish_version = meta.get('publish_version', False)

    _LAMBDA_CONN.update_code_source(
        lambda_name=name,
        s3_bucket=CONFIG.deploy_target_bucket,
        s3_key=key,
        publish_version=publish_version)

    # update lambda layers version
    if meta.get('layers'):
        layers = meta.get('layers')
        updated_layers_arns = [layer_arn
                               for layer_arn, body in context.items()
                               if body.get('resource_name') in layers]
        _LAMBDA_CONN.update_lambda_configuration(lambda_name=name,
                                                 layers=updated_layers_arns)

    # AWS sometimes returns None after function creation, needs for stability
    time.sleep(10)
    response = _LAMBDA_CONN.get_function(name)
    _LOG.debug('Lambda describe result: {0}'.format(response))
    code_sha_256 = response['Configuration']['CodeSha256']
    publish_ver_response = _LAMBDA_CONN.publish_version(
        function_name=name,
        code_sha_256=code_sha_256)
    updated_version = publish_ver_response['Version']
    _LOG.info(
        'Version {0} for lambda {1} published'.format(updated_version, name))

    alias_name = meta.get('alias')
    if alias_name:
        alias = _LAMBDA_CONN.get_alias(function_name=name, name=alias_name)
        if not alias:
            _LAMBDA_CONN.create_alias(
                function_name=name,
                name=alias_name,
                version=updated_version)
            _LOG.info(
                'Alias {0} has been created for lambda {1}'.format(alias_name,
                                                                   name))
        else:
            _LAMBDA_CONN.update_alias(
                function_name=name,
                alias_name=alias_name,
                function_version=updated_version)
            _LOG.info(
                'Alias {0} has been updated for lambda {1}'.format(alias_name,
                                                                   name))


def _is_equal_lambda_layer(new_layer_sha, old_layer_name):
    import base64
    versions = _LAMBDA_CONN.list_lambda_layer_versions(name=old_layer_name)
    for version in versions:
        old_layer = _LAMBDA_CONN.get_lambda_layer_by_arn(
            version['LayerVersionArn'])
        if new_layer_sha == base64.b64decode(
                old_layer['Content']['CodeSha256']):
            return old_layer


def __describe_lambda_by_version(name):
    versions = _LAMBDA_CONN.versions_list(name)
    # find the last created version
    version = max(
        [int(i['Version']) if i['Version'] != '$LATEST' else 0 for i in
         versions])
    if version != 0:
        return _LAMBDA_CONN.get_function(name, str(version))
    else:
        return _LAMBDA_CONN.get_function(name)


@retry
def _create_dynamodb_trigger_from_meta(lambda_name, lambda_arn, role_name,
                                       trigger_meta):
    required_parameters = ['target_table', 'batch_size']
    validate_params(lambda_name, trigger_meta, required_parameters)
    table_name = trigger_meta['target_table']

    if not CONN.dynamodb().is_stream_enabled(table_name):
        CONN.dynamodb().enable_table_stream(table_name)

    stream = CONN.dynamodb().get_table_stream_arn(table_name)
    # TODO support another sub type
    _LAMBDA_CONN.add_event_source(lambda_arn, stream,
                                  trigger_meta['batch_size'],
                                  start_position='LATEST')
    # start_position='LATEST' - in case we did not remove tables before
    _LOG.info('Lambda %s subscribed to dynamodb table %s', lambda_name,
              table_name)


@retry
def _create_sqs_trigger_from_meta(lambda_name, lambda_arn, role_name,
                                  trigger_meta):
    required_parameters = ['target_queue', 'batch_size']
    validate_params(lambda_name, trigger_meta, required_parameters)
    target_queue = trigger_meta['target_queue']

    if not CONN.sqs().get_queue_url(target_queue, CONFIG.account_id):
        _LOG.debug('Queue %s does not exist', target_queue)
        return

    queue_arn = 'arn:aws:sqs:{0}:{1}:{2}'.format(CONFIG.region,
                                                 CONFIG.account_id,
                                                 target_queue)

    _LAMBDA_CONN.add_event_source(lambda_arn, queue_arn,
                                  trigger_meta['batch_size'])
    _LOG.info('Lambda %s subscribed to SQS queue %s', lambda_name,
              target_queue)


@retry
def _create_cloud_watch_trigger_from_meta(lambda_name, lambda_arn, role_name,
                                          trigger_meta):
    required_parameters = ['target_rule']
    validate_params(lambda_name, trigger_meta, required_parameters)
    rule_name = trigger_meta['target_rule']

    rule_arn = CONN.cw_events().get_rule_arn(rule_name)
    CONN.cw_events().add_rule_target(rule_name, lambda_arn)
    _LAMBDA_CONN.add_invocation_permission(lambda_arn, 'events.amazonaws.com',
                                           rule_arn)
    _LOG.info('Lambda %s subscribed to cloudwatch rule %s', lambda_name,
              rule_name)


@retry
def _create_s3_trigger_from_meta(lambda_name, lambda_arn, role_name,
                                 trigger_meta):
    required_parameters = ['target_bucket', 's3_events']
    validate_params(lambda_name, trigger_meta, required_parameters)
    target_bucket = trigger_meta['target_bucket']

    if not _S3_CONN.is_bucket_exists(target_bucket):
        _LOG.error(
            'S3 bucket {0} event source for lambda {1} was not created.'.format(
                target_bucket, lambda_name))
        return
    _LAMBDA_CONN.add_invocation_permission(lambda_arn,
                                           's3.amazonaws.com',
                                           'arn:aws:s3:::{0}'.format(
                                               target_bucket))
    _S3_CONN.configure_event_source_for_lambda(target_bucket, lambda_arn,
                                               trigger_meta['s3_events'],
                                               trigger_meta.get(
                                                   'filter_rules'))
    _LOG.info('Lambda %s subscribed to S3 bucket %s', lambda_name,
              target_bucket)


@retry
def _create_sns_topic_trigger_from_meta(lambda_name, lambda_arn, role_name,
                                        trigger_meta):
    required_params = ['target_topic']
    validate_params(lambda_name, trigger_meta, required_params)
    topic_name = trigger_meta['target_topic']

    region = trigger_meta.get('region')
    create_sns_subscription_for_lambda(lambda_arn, topic_name, region)
    _LOG.info('Lambda %s subscribed to sns topic %s', lambda_name,
              trigger_meta['target_topic'])


@retry
def _create_kinesis_stream_trigger_from_meta(lambda_name, lambda_arn,
                                             role_name, trigger_meta):
    required_parameters = ['target_stream', 'batch_size', 'starting_position']
    validate_params(lambda_name, trigger_meta, required_parameters)

    stream_name = trigger_meta['target_stream']

    stream = CONN.kinesis().get_stream(stream_name)
    stream_arn = stream['StreamDescription']['StreamARN']
    stream_status = stream['StreamDescription']['StreamStatus']
    # additional waiting for stream
    if stream_status != 'ACTIVE':
        _LOG.debug('Kinesis stream %s is not in active state,'
                   ' waiting for activation...', stream_name)
        time.sleep(120)

    # TODO policy should be moved to meta
    policy_name = '{0}KinesisTo{1}Lambda'.format(stream_name, lambda_name)
    policy_document = {
        "Statement": [
            {
                "Effect": "Allow",
                "Action": [
                    "lambda:InvokeFunction"
                ],
                "Resource": [
                    lambda_arn
                ]
            },
            {
                "Action": [
                    "kinesis:DescribeStreams",
                    "kinesis:DescribeStream",
                    "kinesis:ListStreams",
                    "kinesis:GetShardIterator",
                    "Kinesis:GetRecords"
                ],
                "Effect": "Allow",
                "Resource": stream_arn
            }
        ],
        "Version": "2012-10-17"
    }
    CONN.iam().attach_inline_policy(role_name=role_name,
                                    policy_name=policy_name,
                                    policy_document=policy_document)
    _LOG.debug('Inline policy %s is attached to role %s',
               policy_name, role_name)
    _LOG.debug('Waiting for activation policy %s...', policy_name)
    time.sleep(10)

    _add_kinesis_event_source(lambda_arn, stream_arn, trigger_meta)
    _LOG.info('Lambda %s subscribed to kinesis stream %s', lambda_name,
              stream_name)


@retry
def _add_kinesis_event_source(lambda_name, stream_arn, trigger_meta):
    _LAMBDA_CONN.add_event_source(lambda_name, stream_arn,
                                  trigger_meta['batch_size'],
                                  trigger_meta['starting_position'])


CREATE_TRIGGER = {
    'dynamodb_trigger': _create_dynamodb_trigger_from_meta,
    'cloudwatch_rule_trigger': _create_cloud_watch_trigger_from_meta,
    's3_trigger': _create_s3_trigger_from_meta,
    'sns_topic_trigger': _create_sns_topic_trigger_from_meta,
    'kinesis_trigger': _create_kinesis_stream_trigger_from_meta,
    'sqs_trigger': _create_sqs_trigger_from_meta
}


def remove_lambdas(args):
    create_pool(_remove_lambda, args)


@unpack_kwargs
@retry
def _remove_lambda(arn, config):
    lambda_name = config['resource_name']
    try:
        _LAMBDA_CONN.delete_lambda(lambda_name)
        _LAMBDA_CONN.remove_trigger(lambda_name)
        group_names = _CW_LOGS_CONN.get_log_group_names()
        for each in group_names:
            if lambda_name == each.split('/')[-1]:
                _CW_LOGS_CONN.delete_log_group_name(each)
        _LOG.info('Lambda %s was removed.', lambda_name)
    except ClientError as e:
        if e.response['Error']['Code'] == 'ResourceNotFoundException':
            _LOG.warn('Lambda %s is not found', lambda_name)
        else:
            raise e


def create_lambda_layer(args):
    return create_pool(create_lambda_layer_from_meta, args)


@unpack_kwargs
def create_lambda_layer_from_meta(name, meta, context=None):
    """
    :param name:
    :param meta:
    :param context: because of usage in 'update' flow
    :return:
    """
    req_params = ['runtimes', 'file_name']

    validate_params(name, meta, req_params)

    key = meta[S3_PATH_NAME]
    file_name = key.split('/')[-1]
    _S3_CONN.download_file(CONFIG.deploy_target_bucket, key, file_name)
    with open(file_name, 'rb') as file_data:
        file_body = file_data.read()
    import hashlib
    hash_object = hashlib.sha256()
    hash_object.update(file_body)
    existing_version = _is_equal_lambda_layer(hash_object.digest(), name)
    if existing_version:
        existing_layer_arn = existing_version['LayerVersionArn']
        _LOG.info('Layer {} with same content already '
                  'exists in layer version {}.'.format(name,
                                                       existing_layer_arn))
        return {
            existing_layer_arn: build_description_obj(
                response=existing_version, name=name, meta=meta
            )}

    _LOG.debug('Creating lambda layer %s', name)

    args = {'layer_name': name, 'runtimes': meta['runtimes'],
            's3_bucket': CONFIG.deploy_target_bucket,
            's3_key': meta[S3_PATH_NAME]}
    if meta.get('description'):
        args['description'] = meta['description']
    if meta.get('license'):
        args['layer_license'] = meta['license']
    response = _LAMBDA_CONN.create_layer(**args)

    _LOG.info('Lambda Layer {0} version {1} was successfully created'.format(
        name, response['Version']))
    layer_arn = response['LayerArn'] + ':' + str(response['Version'])
    del response['LayerArn']
    return {
        layer_arn: build_description_obj(
            response, name, meta)
    }


def remove_lambda_layers(args):
    return create_pool(_remove_lambda_layers, args)


@unpack_kwargs
@retry
def _remove_lambda_layers(arn, config):
    layer_name = config['resource_name']
    _LOG.info('The latest lambda layer {0} version {1} was found.'.format(
        layer_name, arn.split(':')[-1]))
    layers_list = _LAMBDA_CONN.list_lambda_layer_versions(layer_name)

    try:
        for arn in [layer['LayerVersionArn'] for layer in layers_list]:
            layer_version = arn.split(':')[-1]
            _LAMBDA_CONN.delete_layer(arn)
            _LOG.info('Lambda layer {0} version {1} was removed.'.format(
                layer_name, layer_version))
    except ClientError as e:
        if e.response['Error']['Code'] == 'ResourceNotFoundException':
            _LOG.warn('Lambda Layer {} is not found'.format(layer_name))
        else:
            raise e
